from rest_framework import fields, serializers

from api.utils.serializers.bookmarks import BookmarkedSerializerMixin
from api.utils.serializers.data_refs import DataRefsSerializerMixin
from api.utils.serializers.tags import TagsSerializerMixin
from db.models.jobs import Job, JobStatus
from libs.spec_validation import validate_job_spec_config


class JobStatusSerializer(serializers.ModelSerializer):
    uuid = fields.UUIDField(format='hex', read_only=True)

    class Meta:
        model = JobStatus
        extra_kwargs = {'job': {'read_only': True}}
        exclude = []


class JobSerializer(serializers.ModelSerializer):
    uuid = fields.UUIDField(format='hex', read_only=True)
    user = fields.SerializerMethodField()
    project = fields.SerializerMethodField()
    build_job = fields.SerializerMethodField()
    original = fields.SerializerMethodField()

    class Meta:
        model = Job
        fields = (
            'id',
            'uuid',
            'name',
            'unique_name',
            'pod_id',
            'user',
            'description',
            'created_at',
            'updated_at',
            'last_status',
            'started_at',
            'finished_at',
            'original',
            'cloning_strategy',
            'tags',
            'project',
            'build_job',
        )

    def get_user(self, obj):
        return obj.user.username

    def get_project(self, obj):
        return obj.project.unique_name

    def get_build_job(self, obj):
        return obj.build_job.unique_name if obj.build_job else None

    def get_original(self, obj):
        return obj.original_job.unique_name if obj.original_job else None


class BookmarkedJobSerializer(JobSerializer, BookmarkedSerializerMixin):
    bookmarked_model = 'job'

    class Meta(JobSerializer.Meta):
        fields = JobSerializer.Meta.fields + ('bookmarked',)


class JobDetailSerializer(BookmarkedJobSerializer, TagsSerializerMixin, DataRefsSerializerMixin):
    resources = fields.SerializerMethodField()
    merge = fields.BooleanField(write_only=True, required=False)

    class Meta(BookmarkedJobSerializer.Meta):
        fields = BookmarkedJobSerializer.Meta.fields + (
            'merge',
            'is_clone',
            'original_job',
            'description',
            'readme',
            'config',
            'resources',
            'data_refs',
            'node_scheduled',
        )
        extra_kwargs = {'original_job': {'write_only': True}}

    def get_resources(self, obj):
        return obj.resources.to_dict() if obj.resources else None

    def update(self, instance, validated_data):
        validated_data = self.validated_tags(validated_data=validated_data,
                                             tags=instance.tags)
        validated_data = self.validated_data_refs(validated_data=validated_data,
                                                  data_refs=instance.data_refs)

        return super().update(instance=instance, validated_data=validated_data)


class JobCreateSerializer(serializers.ModelSerializer):
    user = fields.SerializerMethodField()

    class Meta:
        model = Job
        fields = ('id', 'user', 'name', 'description', 'readme', 'data_refs', 'config',)

    def get_user(self, obj):
        return obj.user.username

    def validate_config(self, config):
        """We only validate the config if passed.

        Also we use the JobSpecification to check if this config was
        intended as job.
        """
        validate_job_spec_config(config)
        return config
